/*******************************************************************************
*                                 qdos_c                                       *
*                                 ~~~~~~                                       *
*      QDOS initialization and graphic routines for POV Ray Trace v2.2         *
*       by T. Godefroy 1994. New Locate_File function by J. Grimbert.          *
*                                                                              *
*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <qdos.h>
#include <qptr.h>
#include "frame.h"

#define INFINITE (timeout_t) -1

/* Variables externes (dclares dans povray_c) */

extern unsigned int Options;
extern char DisplayFormat;
extern char PaletteOption;
extern int Display_Started;
extern char *Library_Paths[];
extern int Library_Path_Index;

/* Variables utilises dans qdos_c */

int redirected;
int console_out;
int console_in;
chanid_t stdout_id, stdin_id;
struct QLRECT coord = { 2, 1, 0, 0 };
struct REGS inp_par;
struct REGS out_par;
double Kx, Ky;

/* Nom du programme, version QDOS, copyright */

char _prog_name[]   = "POV";
char _version[24]   = "v2.2a";
char _copyright[32] = "QDOS/SMSQ port";

/* Paramtres de la fentre d'affichage */

void (*_consetup)() = NULL;

struct WINDOWDEF _condetails  = { 208, 1, 0, 7, 512, 256, 0, 0};
struct QLRECT window_size     = { 512, 242, 0, 14 };
struct QLRECT win_outline     = { 512, 256, 0,  0 };
struct WM_wsiz max_size       = { 512, 256, 0,  0 };

/* Prompt lors d'un exit() */

char *_endmsg = "\nPress a key to exit.";

/* Besoins et limitations en mmoire */

long _stack   =   16L * 1024L;
long _mneed   =   16L * 1024L;
long _memmax  = 9999L * 1024L;
long _memincr =   16L * 1024L;
long _memqdos =   24L * 1024L;

/* Routine de lecture/changement du mode d'affichage:
     si m = -1        : retourne le mode d'affichage actuel;
     si m = 4, 8 ou 12: change le mode et retourne TRUE si russi, FALSE sinon.
*/

short do_mode(m)
short m;
{
    short a = -1;
    short n =  m;

    mt_dmode(&n, &a);

    if (m == -1)
        return(n);

    a = -1;
    n = -1;
    mt_dmode(&n, &a);

    return((short) n == m);
}

/* Routine de dessin de la fentre d'affichage (si non redirige) */

void draw_stdout()
{
    short m;
    char *c;

    if (!redirected) {

        /* Passer en mode 4 ou rester en mode 2 (SMS2) */

        m = do_mode(-1);
        if (m != 4 && m != 2)
            do_mode(4);

        /* Redessiner la fentre de stdout */

        iop_outl(stdout_id, INFINITE, 0, 0, 0, &win_outline);
        fflush(stdout);
        sd_wdef(stdout_id, INFINITE, 0, 0, &win_outline);
        fflush(stdout);
        consetup_title(stdout_id, &_condetails);
        fflush(stdout);
        sd_setin(stdout_id, INFINITE, 4);
        fflush(stdout);
        fflush(stdin);
        sd_cure (stdin_id, INFINITE);
        fflush(stdin);
        io_fbyte(stdin_id, 0, c);
        fflush(stdin);
    }
}

/* Routine d'initialisation du programme */

void qdos_init()
{
    char *s = _version;

    /* Mettre la date de compilation derrire le numro de version */

    strcat(s, " (");
    strcat(s, __DATE__);
    strcat(s, ")");

    /* Stocker les identificateurs de canaux QDOS de stdout */

    stdout_id = fgetchid(stdout);
    stdin_id  = fgetchid(stdin);

    /* Flags de redirection et de type de priphrique */

    redirected  = isnoclose(fileno(stdout));
    console_out = iscon(stdout_id, INFINITE);
    console_in  = iscon(stdin_id, INFINITE);

    /* Chercher la taille maximale de l'cran (si l'EP est prsent !) */

    iop_flim(stdout_id, INFINITE, &max_size);

    draw_stdout();
}

/* Routine de sortie du programme */

void qdos_exit()
{
/*  _endmsg = NULL; */
}

/* Routine de test pour sortie immdiate du programme: active si option "+x" */

void test_abortion()
{
    char *c1, c2;
    int err;

    *c1 = 0;
    c2  = 0;

    if (console_in && (Options & EXITENABLE)) {

        while (io_fbyte(stdin_id, 0, c1) != -1)
            c2 = *c1;

        /* si caractre CTRL Q, quitter */

        if (c2 == '\021') {

            if (Display_Started)
                draw_stdout();

            printf("\n\nAborted !\n");
            exit(-1);
        }
    }
}

/* Routine de test permettant de continuer les calculs sans tenter d'crire
   les messages de l'option "+v" sur une fentre gele, en arrire plan, ou
   pendant un affichage graphique.
*/

int isnotlocked()
{
    if (console_out) {
        if (Display_Started)
            return(FALSE);
        else
            return(sd_cure(stdout_id, 0) == 0);
    }
    else
        return(TRUE);
}

/* Routine d'initialisation de l'affichage graphique (option "+dN"):
   si N=0, couleurs "vraies" (dominantes rouges/vertes/bleues respectes);
   si N=1, niveaux de gris, chelle linaire;
   si N=2, niveaux de gris, chelle linaire, faible intensit (0  50%);
   si N=3, niveaux de gris, chelle linaire, intensit moyenne (25%  75%);
   si N=4, niveaux de gris, chelle linaire, forte intensit (50%  100%);
   si N=5, niveaux de gris, chelle logarithmique de base 2 (0  100%).
*/

void display_init(Width, Height)
int Width, Height;
{
    struct QLRECT winsiz;

    if (redirected) {

        /* Il est hors de question de modifier une fentre redirige (le
           job auquel elle appartient pourrait bien s'en offusquer ! */

        printf("\nGraphic displaying on a redirected window is not allowed !\n");
        exit(-1);

    } else {

        /* Vrification de la validit du format demand */

        switch (DisplayFormat) {
            case '0':
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
                break;
            default:
                printf("\n\nBad display format !\n");
                exit(-1);
        }

        /* Mettre l'ordinateur dans le mode possdant le plus de couleurs
          (16, 8 ou 4 selon que l'on est sur Thor XVI, Thor 8/QL ou ATARI-QL)
        */

        sd_bordr(stdout_id, INFINITE, 0, 0);
        fflush(stdout);

        if (!do_mode(12))
            if (!do_mode(8))
                do_mode(4);

        /* Coefficients d'chelle par dfaut */

        Kx = 2.0;
        Ky = 1.0;

        /* Dimensionner la fentre d'affichage */

        winsiz.q_width  = 2 * Width;
        winsiz.q_height = Height;
        winsiz.q_x      = 0;
        winsiz.q_y      = 0;

        if (sd_wdef(stdout_id, INFINITE, 0, 0, &winsiz) != 0) {

            winsiz.q_width  = max_size.xsize;
            winsiz.q_height = max_size.ysize;
            winsiz.q_x      = max_size.xorg;
            winsiz.q_y      = max_size.yorg;

            sd_wdef(stdout_id, INFINITE, 0, 0, &winsiz);
            iop_outl(stdout_id, INFINITE, 0, 0, 0, &max_size);

            if ((2 * Width) > max_size.xsize)
                Kx = ((double) max_size.xsize) / ((double) Width);

            if (Height > max_size.ysize)
                Ky = ((double) max_size.ysize) / ((double) Height);
        } else {

            iop_outl(stdout_id, INFINITE, 0, 0, 0, &winsiz);
        }

        sd_clear(stdout_id, INFINITE);

        /* Inhiber l'affichage du curseur */

        sd_curs(stdout_id, INFINITE);
        fflush(stdout);

        /* Initialiser certains paramtres pour display_plot() */

        inp_par.D0 = 46L;
        inp_par.D3 = -1L;
        inp_par.A0 = (addreg_t) stdout_id;
        inp_par.A1 = (addreg_t) &coord;
    }
}

/* Routine d'affichage d'un point */

void display_plot(x, y, Red, Green, Blue)
int x, y;
unsigned char Red, Green, Blue;
{
    int color, intensity, R, G, B;

    R = (int) Red;
    G = (int) Green;
    B = (int) Blue;

    intensity = (B + R + G) / 3;

    switch (DisplayFormat) {

        case '0':   /* affichage en couleurs "vraies" */
            if (intensity == 0)
                color = 0;
            else {
                color  = ((B / intensity) >= 1);
                color += ((R / intensity) >= 1) * 2;
                color += ((G / intensity) >= 1) * 4;
                color += (intensity < 128) * 32768;         /* pour MODE 12 */
                }
            break;

        case '1':   /* niveaux de gris, chelle linaire (0  100%) */
            color = intensity / 16;
            color = color / 2 + 32768 * (1 - color % 2);
            break;

        case '2':   /* niveaux de gris, chelle linaire (0  50%) */
            color = intensity / 8;
            if (color > 15)
                color = 15;
            color = color / 2 + 32768 * (1 - color % 2);
            break;

        case '3':   /* niveaux de gris, chelle linaire (25  75%) */
            color = (intensity - 64) / 8;
            if (color < 0)
                color = 0;
            if (color > 15)
                color = 15;
            color = color / 2 + 32768 * (1 - color % 2);
            break;

        case '4':   /* niveaux de gris, chelle linaire (50  100%) */
            color = (intensity - 128) / 8;
            if (color < 0)
                color = 0;
            color = color / 2 + 32768 * (1 - color % 2);
            break;

        case '5':   /* niveaux de gris, chelle logarithmique (0  100%) */
            if (intensity == 0)
                color = 0;
            else {
                color = (int)(2.0 * log((double)(intensity)) / log(2.0));
                color = color / 2 + 32768 * (1 - color % 2);
            }
    }

    /* Pour afficher un point on utilise TRAP #3, D0=$2E: trac d'un rectangle
       de couleur (ici taille rectangle=2x1 pixels) en coordonnes graphiques.
       Ici "qdos3" est utilis en lieu et place de "sd_fill" pour permettre le
       passage du paramtre "color" sous forme d'un mot long (pour affichage 16
       couleurs sur Thor XVI en MODE 12).
    */

    inp_par.D1 = color;

    coord.q_x  = (int) (Kx * ((double) x));
    coord.q_y  = (int) (Ky * ((double) y));

    qdos3(&inp_par, &out_par);
    fflush(stdout);
}

/* Routine appele en fin d'affichage avant de refermer la fentre graphique */

void display_finished()
{
    if (Display_Started) {
        beep(5000, 30);

        if (console_in)
            getchar();
    }
}

/* Routine de fermeture de la fentre graphique */

void display_close()
{
    draw_stdout();
}

/* Traitement des erreurs de calcul (overflow, sorties des limites...) */

int matherr (x)
struct exception *x;
{
    switch(x->type) {
        case DOMAIN:
        case OVERFLOW:
            x->retval = 1.0e17;
            break;

        case SING:
        case UNDERFLOW:
            x->retval = 0.0;
            break;

        default:
            break;
    }
    return(1);
}

/* Routine amliore de recherche de fichier: Merci Jrme Grimbert ! */

/* Find a file in the search path. */

FILE *Locate_File(filename, mode)
char *filename, *mode;
{
    FILE *f;
    int i;
    char pathname[FILE_NAME_LENGTH];
    char qdosfilename[FILE_NAME_LENGTH];
    unsigned char length, qdos;          /* don't use very large file name... */


    /* Check the current directory first. */

    if ((f = fopen (filename, mode)) != NULL)
        return (f);

    for (i = 0; i < Library_Path_Index; i++) {

        strcpy(pathname, Library_Paths[i]);

        if (FILENAME_SEPARATOR != NULL)
            strcat(pathname, FILENAME_SEPARATOR);

        strcat (pathname, filename);

        if ((f = fopen(pathname, mode)) != NULL)
            return (f);
    }

    /* would you mind to look at the last . as a _ on the file system ?
       This avoid user a lot of renaming of file name and allow to keep new
       scenario compatible
    */

    strcpy(qdosfilename, filename );
    length = strlen(qdosfilename);

    for (qdos = length; qdos; qdos--)
        if (qdosfilename[qdos] == '.') {

            qdosfilename[qdos] = '_';
            qdos=0;
        }

    if ((f = fopen(qdosfilename, mode)) != NULL)
        return (f);

    for (i = 0; i < Library_Path_Index; i++) {

        strcpy(pathname, Library_Paths[i]);

        if (FILENAME_SEPARATOR != NULL)
            strcat(pathname, FILENAME_SEPARATOR);

        strcat(pathname, qdosfilename);

        if ((f = fopen(pathname, mode)) != NULL)
            return(f);
    }

    return (NULL);
}

/* FIN */
